@@ -86,7 +86,7 @@ typedef struct {
    float scale;			/* old scale value */
 } scaleinfo;
 
-Boolean undo_collect = FALSE;
+u_char undo_collect = (u_char)0;
 
 /*----------------------------------------------------------------------*/
 /* Externally declared variables					*/
@@ -200,6 +200,40 @@ short *regen_selection(objinstptr thisinst, uselection *srec)
       return slist;
 }
 
+/*----------------------------------------------------------------------*/
+/* Use the selection list in the undo record to reorder parts in the	*/
+/* object's part list.  Invert the selection list and return it.	*/
+/*----------------------------------------------------------------------*/
+
+void reorder_selection(Undoptr thisrecord)
+{
+   short *slist, *newlist, snum, i;
+   genericptr *pgen, *plist, egen;
+   objinstptr thisinst = thisrecord->thisinst;
+   objectptr thisobj = thisinst->thisobject;
+
+   snum = (short)thisrecord->idata;
+   slist = (short *)thisrecord->undodata;
+   plist = (genericptr *)malloc(snum * sizeof(genericptr));
+   newlist = (short *)malloc(snum * sizeof(short));
+
+   i = 0;
+   for (pgen = plist; pgen < plist + snum; pgen++) {
+      egen = *(thisobj->plist + i);
+      *(plist + *(slist + i)) = egen;
+      i++;
+   }
+   i = 0;
+   for (pgen = plist; pgen < plist + snum; pgen++) {
+      *(thisobj->plist + i) = *pgen;
+      *(newlist + *(slist + i)) = i;
+      i++;
+   }
+   free(plist);
+   free(thisrecord->undodata);
+   thisrecord->undodata = (char *)newlist;
+}
+
 /*----------------------------------------------------------------------*/
 /* free_editelement ---							*/
 /*									*/
@@ -424,7 +458,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
    else
       newrecord->idx = 1;
 
-   if (mode == UNDO_MORE || undo_collect == TRUE)
+   if (mode == UNDO_MORE || undo_collect > (u_char)0)
       newrecord->idx = -newrecord->idx;
    
    xobjs.undostack = newrecord;
@@ -556,6 +590,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 
       case XCF_ChangeStyle:
       case XCF_Justify:
+      case XCF_Color:
 	 /* 2 args:							*/
 	 /*	egen = element that was changed (with new value)	*/
 	 /*	oval = old style value					*/
@@ -613,6 +648,15 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 	 /* Fprintf(stdout, "Undo: registered move of delta (%d, %d)\n",
 			deltax, deltay); */
 	 break;
+
+       case XCF_Reorder:
+	 /* 2 args:							*/
+	 /*	slist = "before" order of elements			*/
+	 /*	snum = number of elements in list (= # parts)		*/
+	 slist = va_arg(args, short *);
+	 snum = va_arg(args, int);
+	 newrecord->undodata = (char *)slist;
+	 newrecord->idata = snum;
    }
 
    va_end(args);
@@ -703,8 +747,7 @@ short undo_one_action()
 	    break;
 	 }
 	 else {
-	    /* XXX egen is uninitialized! */
-	    delete_one_element(thisrecord->thisinst, egen);
+	    delete_one_element(thisrecord->thisinst, (genericptr)thisinst);
 	 }
 
 	 /* Put back all the parts */
@@ -804,9 +847,10 @@ short undo_one_action()
 	    case OBJINST: {
 	       XPoint tmppt;
 	       objinstptr einst = (objinstptr)(erec->element);
+	       // Swap instance and saved positions.
 	       tmppt = einst->position;
 	       einst->position = erec->save.instpos;
-	       erec->save.instpos = ((objinstptr)egen)->position;
+	       erec->save.instpos = tmppt;
 	       drawarea(areawin->area, NULL, NULL);
 	    }  break;
 	    case POLYGON: {
@@ -876,6 +920,7 @@ short undo_one_action()
 	 break;
 
       case XCF_Copy:
+	 clearselects_noundo();
 	 srec = (uselection *)thisrecord->undodata;
 	 slist = regen_selection(thisrecord->thisinst, srec);
 	 thisobj = delete_element(thisrecord->thisinst, slist,
@@ -886,6 +931,7 @@ short undo_one_action()
 	 /* Revert selection to previously selected */
 	 select_previous(thisrecord);
 	 drawarea(areawin->area, NULL, NULL);
+	 draw_all_selected();
 	 break;
 
       case XCF_ChangeStyle:
@@ -917,6 +963,35 @@ short undo_one_action()
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
+      case XCF_Color:
+	 /* Color changes */
+	 egen = (genericptr)thisrecord->undodata;
+	 snum = thisrecord->idata;
+	 switch(egen->type) {
+	    case PATH:
+	       thispath = (pathptr)(thisrecord->undodata);
+	       thisrecord->idata = thispath->color;
+	       thispath->color = snum;
+	       break;
+	    case POLYGON:
+	       thispoly = (polyptr)(thisrecord->undodata);
+	       thisrecord->idata = thispoly->color;
+	       thispoly->color = snum;
+	       break;
+	    case ARC:
+	       thisarc = (arcptr)(thisrecord->undodata);
+	       thisrecord->idata = thisarc->color;
+	       thisarc->color = snum;
+	       break;
+	    case SPLINE:
+	       thisspline = (splineptr)(thisrecord->undodata);
+	       thisrecord->idata = thisspline->color;
+	       thisspline->color = snum;
+	       break;
+	 }
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
       case XCF_Rescale:
 	 escale = (scaleinfo *)thisrecord->undodata;
 	 egen = escale->element;
@@ -983,6 +1058,12 @@ short undo_one_action()
 	 placeselects(-(delta->x), -(delta->y), NULL);
 	 reset_cycles();
 	 drawarea(areawin->area, NULL, NULL);
+	 draw_all_selected();
+	 break;
+
+       case XCF_Reorder:
+	 reorder_selection(thisrecord);
+	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
       default:
@@ -1015,12 +1096,12 @@ short undo_one_action()
 /*	Complete a possibly incomplete undo series by forcing the	*/
 /*	topmost entry to have a positive index.				*/
 /*	Note that for "undo series start|end" to work, undo_collect	*/
-/*	must be set to FALSE prior to calling undo_finish_series().	*/
+/*	must be set to 0 prior to calling undo_finish_series().		*/
 /*----------------------------------------------------------------------*/
 
 void undo_finish_series()
 {
-   if (undo_collect == FALSE)
+   if (undo_collect == (u_char)0)
       if (xobjs.undostack && xobjs.undostack->idx < 0)
          xobjs.undostack->idx = -xobjs.undostack->idx;
 }
@@ -1032,9 +1113,9 @@ void undo_finish_series()
 
 void undo_action()
 {
-   short idx = undo_one_action(False);
+   short idx = undo_one_action();
    while (xobjs.undostack && xobjs.undostack->idx == idx)
-      undo_one_action(False);
+      undo_one_action();
 }
 
 /*----------------------------------------------------------------------*/
@@ -1187,9 +1268,10 @@ short redo_one_action()
 	    case OBJINST: {
 	       XPoint tmppt;
 	       objinstptr einst = (objinstptr)(erec->element);
+	       // Swap instance position and saved position
 	       tmppt = einst->position;
 	       einst->position = erec->save.instpos;
-	       erec->save.instpos = ((objinstptr)egen)->position;
+	       erec->save.instpos = tmppt;
 	       drawarea(areawin->area, NULL, NULL);
 	    }  break;
 	    case ARC: {
@@ -1305,6 +1387,35 @@ short redo_one_action()
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
+      case XCF_Color:
+	 /* Color changes */
+	 egen = (genericptr)thisrecord->undodata;
+	 snum = thisrecord->idata;
+	 switch(egen->type) {
+	    case PATH:
+	       thispath = (pathptr)(thisrecord->undodata);
+	       thisrecord->idata = thispath->color;
+	       thispath->color = snum;
+	       break;
+	    case POLYGON:
+	       thispoly = (polyptr)(thisrecord->undodata);
+	       thisrecord->idata = thispoly->color;
+	       thispoly->color = snum;
+	       break;
+	    case ARC:
+	       thisarc = (arcptr)(thisrecord->undodata);
+	       thisrecord->idata = thisarc->color;
+	       thisarc->color = snum;
+	       break;
+	    case SPLINE:
+	       thisspline = (splineptr)(thisrecord->undodata);
+	       thisrecord->idata = thisspline->color;
+	       thisspline->color = snum;
+	       break;
+	 }
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
       case XCF_Rescale:
 	 escale = (scaleinfo *)thisrecord->undodata;
 	 egen = escale->element;
@@ -1363,6 +1474,11 @@ short redo_one_action()
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
+      case XCF_Reorder:
+	 reorder_selection(thisrecord);
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
       default:
 	 Fprintf(stderr, "Undo not implemented for this action!\n");
 	 break;
@@ -1503,6 +1619,7 @@ void free_undo_data(Undoptr thisrecord, u_char mode)
       case XCF_Push:
       case XCF_ChangeStyle:
       case XCF_Justify:
+      case XCF_Color:
 	 /* Do nothing --- undodata points to a valid element */
 	 break;
 
